import 'dart:async';
import 'dart:convert';

import 'package:flutter_test/flutter_test.dart';
import 'package:hydrated/hydrated.dart';
import 'package:shared_preferences/shared_preferences.dart';

void main() {
  setUp(() {
    SharedPreferences.setMockInitialValues({
      "flutter.prefs": true,
      "flutter.int": 1,
      "flutter.double": 1.1,
      "flutter.bool": true,
      "flutter.String": "first",
      "flutter.List<String>": ["a", "b"],
      "flutter.SerializedClass": '{"value":true,"count":42}'
    });
  });

  test('Shared Preferences set mock initial values', () async {
    final prefs = await SharedPreferences.getInstance();

    final value = prefs.getBool("prefs");
    expect(value, isTrue);
  });

  group('HydratedSubject', () {
    group('correctly handles data type', () {
      test('int', () async {
        await testHydrated<int>("int", 1, 2);
      });

      test('double', () async {
        await testHydrated<double>("double", 1.1, 2.2);
      });

      test('bool', () async {
        await testHydrated<bool>("bool", true, false);
      });

      test('String', () async {
        await testHydrated<String>("String", "first", "second");
      });

      test('List<String>', () async {
        testHydrated<List<String>>("List<String>", ["a", "b"], ["c", "d"]);
      });

      test('SerializedClass', () async {
        final completer = Completer();

        final subject = HydratedSubject<SerializedClass>(
          "SerializedClass",
          hydrate: (s) => SerializedClass.fromJSON(s),
          persist: (c) => c.toJSON(),
          onHydrate: () => completer.complete(),
        );

        final second = SerializedClass(false, 42);

        /// null before hydrate
        expect(subject.valueOrNull, isNull);

        /// properly hydrates
        await completer.future;
        expect(subject.value.value, isTrue);
        expect(subject.value.count, equals(42));

        /// add values
        subject.add(second);
        expect(subject.value.value, isFalse);
        expect(subject.value.count, equals(42));

        /// check value in store
        final prefs = await SharedPreferences.getInstance();
        expect(prefs.get(subject.key), equals('{"value":false,"count":42}'));

        /// clean up
        subject.close();
      });
    });
  });

  test('HydratedSubject emits latest value into the new listener', () async {
    final subject = HydratedSubject<SerializedClass>(
      "SerializedClass",
      hydrate: (s) => SerializedClass.fromJSON(s),
      persist: (c) => c.toJSON(),
    );

    await subject.first;

    final expectation = expectLater(
        subject.stream,
        emitsInOrder([
          isA<SerializedClass>().having((c) => c.value, 'value', isTrue),
          isA<SerializedClass>().having((c) => c.value, 'value', isFalse),
        ]));

    final second = SerializedClass(false, 42);
    subject.add(second);

    await expectation;
    subject.close();
  });
}

/// An example of a class that serializes to and from a string
class SerializedClass {
  final bool value;
  final int count;

  SerializedClass(this.value, this.count);

  factory SerializedClass.fromJSON(String s) {
    final map = jsonDecode(s);

    return SerializedClass(
      map['value'],
      map['count'],
    );
  }

  String toJSON() => jsonEncode({
        'value': this.value,
        'count': this.count,
      });
}

/// The test procedure for a HydratedSubject
Future<void> testHydrated<T>(
  String key,
  T first,
  T second,
) async {
  final completer = Completer();

  final subject = HydratedSubject<T>(
    key,
    onHydrate: () => completer.complete(),
  );

  /// null before hydrate
  expect(subject.valueOrNull, isNull);
  expect(subject.hasValue, isFalse);

  /// properly hydrates
  await completer.future;
  expect(subject.value, equals(first));
  expect(subject.hasValue, isTrue);

  /// add values
  subject.add(second);
  expect(subject.value, equals(second));
  expect(subject.hasValue, isTrue);

  /// check value in store
  final prefs = await SharedPreferences.getInstance();
  expect(prefs.get(subject.key), equals(second));

  /// clean up
  subject.close();
}
